﻿// Copyright ©2011 zundairo http://zundairo.net/
using System;
using System.Collections.Generic;
using ZundaFramework.Util;

namespace ZundaFramework.Neurons {
    /// <summary>
    /// 微分方程式の右辺を定義した関数です．
    /// </summary>
    /// <returns></returns>
    public delegate double ODENFunction();

    /// <summary>
    /// 刺激入力の式を定義した関数です．
    /// </summary>
    /// <param name="time">シミュレーション時間</param>
    /// <returns>刺激電流</returns>
    public delegate double StimulusCurrentFunction(double time);

    /// <summary>
    /// 膜電位微分方程式に追加される電流の式です．
    /// </summary>
    /// <returns></returns>
    public delegate double CurrentFunction();

    /// <summary>
    /// Neuronインスタンスが更新されたときに発生するイベントです．
    /// </summary>
    public delegate void StatusEventHandler();

    /// <summary>
    /// 神経細胞・樹状突起・軸索，共通の抽象クラスです．
    /// </summary>
    //泥臭い処理は全部ここに詰め込む
    public abstract class Neuron : IEnumerable<double> {
        #region 神経細胞共通の何か
        private double capacitance = 1.0;
        private double temperature = 1.0;

        /// <summary>
        /// 細胞・軸索・樹状突起の膜電位
        /// </summary>
        public double Potential {
            set { odeParams[0] = value; }
            get { return odeParams[0]; }
        }

        /// <summary>
        /// 膜容量
        /// </summary>
        public double Capacitance {
            set { capacitance = value; }
            get { return capacitance; }
        }

        /// <summary>
        /// 温度定数
        /// </summary>
        public double Temperature {
            set { temperature = value; }
            get { return temperature; }
        }
        #endregion

        private List<string> odeNames = new List<string>();
        private List<double> odeParams = new List<double>();
        private List<ODENFunction> odeNFunctions = new List<ODENFunction>();
        private StimulusCurrentFunction stimulusCurrentFunction = delegate { return 0.0; };
        private List<CurrentFunction> currents = new List<CurrentFunction>();

        /// <summary>
        /// 微分計算によってNeuronの状態が変化した場合のイベントです．
        /// </summary>
        public event StatusEventHandler OnUpdateStatus;

        /// <summary>
        /// インスタンスが持つ微分が必要な変数の数を返します．
        /// </summary>
        public int Length { get { return odeParams.Count; } }

        /// <summary>
        /// 微分方程式の名前コレクションを返します．
        /// </summary>
        public List<string> ODENames { get { return odeNames; } }

        /// <summary>
        /// 微分方程式の解を返す関数のコレクションを返します．
        /// </summary>
        public List<ODENFunction> ODEFunctions { get { return odeNFunctions; } }

        /// <summary>
        /// Neuronインスタンスに微分方程式を追加します．
        /// </summary>
        /// <param name="ode">微分方程式</param>
        /// <param name="value">初期値</param>
        /// <param name="name">変数の名前</param>
        public void AddODE(ODENFunction ode, double value, string name) {
            odeNFunctions.Add(ode);
            odeParams.Add(value);
            odeNames.Add(name);
        }

        /// <summary>
        /// 刺激入力を定義する関数の取得と設定をします．
        /// </summary>
        public StimulusCurrentFunction StimulusCurrentFunction {
            set { stimulusCurrentFunction = value; }
            get { return stimulusCurrentFunction; }
        }

        /// <summary>
        /// インスタンスが持つ微分が必要な変数の取得と設定をします．
        /// </summary>
        /// <param name="i">インデックス(最大this.length)</param>
        /// <returns>変数</returns>
        public double this[int i] {
            get { return odeParams[i]; }
            set { odeParams[i] = value; }
        }

        /// <summary>
        /// 刺激入力電流を取得します．
        /// </summary>
        public double GetStimulusCurrent() {
            return stimulusCurrentFunction(SimulationTime.Time);
        }

        /// <summary>
        /// 膜電位微分方程式に電流を定義する関数を追加します．
        /// </summary>
        /// <param name="current">追加する電流の式</param>
        public void AddCurrent(CurrentFunction current) {
            currents.Add(current);
        }

        /// <summary>
        /// 膜電位微分方程式に電流を定義する関数を追加します．
        /// </summary>
        /// <param name="funcs">追加する電流の式のコレクション</param>
        public void AddCurrent(IEnumerable<CurrentFunction> funcs) {
            currents.AddRange(funcs);
        }

        /// <summary>
        /// 膜電位微分方程式に電流を定義する関数を追加します．
        /// </summary>
        /// <param name="funcs">追加する電流の式</param>
        public void AddCurrent(params CurrentFunction[] funcs) {
            currents.AddRange(funcs);
        }

        /// <summary>
        /// 微分方程式：膜電位v
        /// [dV / dt = 温度定数 * (各電流の和) / 膜容量] によって計算されます．
        /// </summary>
        /// <returns></returns>
        public virtual double ODEPotential() {
            double totalCurrents = 0.0;
            foreach(CurrentFunction current in currents) {
                totalCurrents += current();
            }
            return temperature * totalCurrents / capacitance;
        }

        /// <summary>
        /// Neuronの状態が変化したとき，ODEのCalcから呼び出します．
        /// </summary>
        public void Update() {
            if(OnUpdateStatus != null) {
                OnUpdateStatus();
            }
        }


        #region IEnumerable<double> メンバ

        /// <summary>
        /// NeuronのEnumeratorを取得します．
        /// </summary>
        /// <returns></returns>
        public IEnumerator<double> GetEnumerator() {
            for(int i = 0; i < Length; i++) {
                yield return this[i];
            }
        }

        #endregion

        #region IEnumerable メンバ

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
            throw new Exception("The method or operation is not implemented.");
        }

        #endregion
    }
}
